@@ -157,6 +157,8 @@ class DependenciesHelper:
             elif isinstance(obj, dependencies.ExternalDependency) and obj.name == 'threads':
                 pass
             elif isinstance(obj, dependencies.InternalDependency) and all(lib.get_id() in self.metadata for lib in obj.libraries):
+                FeatureNew.single_use('pkgconfig.generate requirement from internal dependency', '1.9.0',
+                                      self.state.subproject, location=self.state.current_node)
                 # Ensure BothLibraries are resolved:
                 if self.pub_libs and isinstance(self.pub_libs[0], build.StaticLibrary):
                     obj = obj.get_as_static(recursive=True)
@@ -166,8 +168,10 @@ class DependenciesHelper:
                     processed_reqs.append(self.metadata[lib.get_id()].filebase)
             else:
                 raise mesonlib.MesonException('requires argument not a string, '
-                                              'library with pkgconfig-generated file '
-                                              f'or pkgconfig-dependency object, got {obj!r}')
+                                              'library with pkgconfig-generated file, '
+                                              'pkgconfig-dependency object, or '
+                                              'internal-dependency object with '
+                                              f'pkgconfig-generated file, got {obj!r}')
         return processed_reqs
 
     def add_cflags(self, cflags: T.List[str]) -> None:
@@ -295,11 +299,20 @@ class DependenciesHelper:
             self.version_reqs[name].update(version_reqs)
 
     def split_version_req(self, s: str) -> T.Tuple[str, T.Optional[str]]:
+        stripped_str = s.strip()
+        if not stripped_str:
+            raise mesonlib.MesonException(f'required dependency must not be empty, "{s}" was provided.')
         for op in ['>=', '<=', '!=', '==', '=', '>', '<']:
-            pos = s.find(op)
-            if pos > 0:
-                return s[0:pos].strip(), s[pos:].strip()
-        return s, None
+            pos = stripped_str.find(op)
+            if pos < 0:
+                continue
+            if pos == 0:
+                raise mesonlib.MesonException(f'required versioned dependency "{s}" is missing the dependency\'s name.')
+            stripped_str, version = stripped_str[0:pos].strip(), stripped_str[pos:].strip()
+            if not stripped_str:
+                raise mesonlib.MesonException(f'required versioned dependency "{s}" is missing the dependency\'s name.')
+            return stripped_str, version
+        return stripped_str, None
 
     def format_vreq(self, vreq: str) -> str:
         # vreq are '>=1.0' and pkgconfig wants '>= 1.0'
